@@ -1,12 +1,13 @@
 package App::perlmv;
 BEGIN {
-  $App::perlmv::VERSION = '0.37';
+  $App::perlmv::VERSION = '0.38';
 }
 # ABSTRACT: Rename files using Perl code.
 
 use strict;
 use warnings;
 use Cwd qw(abs_path getcwd);
+#use Data::Dump qw(dump);
 use File::Copy;
 use File::Find;
 use File::Path qw(make_path);
@@ -28,12 +29,13 @@ sub new {
             $homedir = File::HomeDir->my_home;
         };
 
-        $homedir ||= $ENV{'HOME'};
+        $homedir //= $ENV{'HOME'};
 
         die "FATAL: Can't determine home directory\n" unless $homedir;
     }
 
     my $self = {
+        codes           => [],
         dry_run         => 0,
         homedir         => $homedir,
         overwrite       => 0,
@@ -54,9 +56,10 @@ sub parse_opts {
 
     GetOptions(
         'c|check'         => \$self->{ 'check'         },
-        'e|execute=s'     => \$self->{ 'code'          },
         'D|delete=s'      => \$self->{ 'delete'        },
         'd|dry-run'       => \$self->{ 'dry_run'       },
+        'e|eval=s'        =>  $self->{ 'codes'         },
+        'h|help'          => sub { $self->print_help() },
         'l|list'          => \$self->{ 'list'          },
         'M|mode=s'        => \$self->{ 'mode'          },
         'o|overwrite'     => \$self->{ 'overwrite'     },
@@ -68,8 +71,9 @@ sub parse_opts {
         'w|write=s'       => \$self->{ 'write'         },
         'f|files'         => sub { $self->{ 'process_dir'    } = 0 },
         'S|no-symlinks'   => sub { $self->{ 'process_symlink'} = 0 },
-        'h|help'          => sub { $self->print_help()             },
         'V|version'       => sub { $self->print_version()          },
+        # we use \scalar to differentiate between -x and -e
+        'x|execute=s'     => sub { push @{$self->{'codes'}}, \$_[1]},
         '<>'              => sub { $self->parse_extra_opts(@_)     },
     ) or $self->print_help();
 }
@@ -99,7 +103,7 @@ sub run {
         'rename';
 
     $self->{'dry_run'} and $self->{'verbose'}++;
-    $self->{'mode'} ||= $default_mode;
+    $self->{'mode'} //= $default_mode;
 
     if ( $self->{'list'} ) {
         $self->load_scriptlets();
@@ -118,7 +122,9 @@ sub run {
     }
 
     if ( $self->{'write'} ) {
-        $self->store_scriptlet( $self->{'write'}, $self->{'code'} );
+        die "Please specify code of scriptlet" unless @{ $self->{'codes'} }
+            && !ref( $self->{'codes'}[0] );
+        $self->store_scriptlet( $self->{'write'}, $self->{'codes'}[0] );
         exit 0;
     }
 
@@ -127,13 +133,15 @@ sub run {
         exit 0;
     }
 
-    unless ($self->{'code'}) {
-        die 'FATAL: Must specify code (-e) or scriptlet name (first argument)'
+    unless (@{ $self->{'codes'} }) {
+        die 'FATAL: Must specify code (-e) or scriptlet name (-x/first argument)'
             unless $self->{'items'};
-        $self->{'code'} =
-            $self->load_scriptlet( scalar shift @{ $self->{'items'} } );
+        push @{ $self->{'codes'} }, \( scalar shift @{ $self->{'items'} } );
+    }
+    # convert all scriptlet names into their code
+    for (@{ $self->{'codes'} }) {
+        $_ = $self->load_scriptlet($$_) if ref($_);
     }
-    exit 0 if $self->{'check'};
 
     die "FATAL: Please specify some files in arguments\n"
         unless $self->{'items'};
@@ -153,40 +161,56 @@ Rename files using Perl code.
 
 Usage:
 
+ # Show help
  perlmv -h
 
+ # Execute a single scriptlet
  perlmv [options] <scriptlet> <file...>
+
+ # Execute code from command line
  perlmv [options] -e <code> <file...>
 
+ # Execute multiple scriptlets/command-line codes
+ perlmv [options] [ -x <scriptlet> | -e <code> ]+ <file...>
+
+ # Create a new scriptlet
  perlmv -e <code> -w <name>
+
+ # List available scriptlets
  perlmv -l
+
+ # Show source code of a scriptlet
  perlmv -s <name>
- perlmv -D <name>
+
+ # Delete scriptlet
+ perlmv -d <name>
 
 Options:
 
  -c  (--compile) Only test compile code, do not run it on the arguments
- -e <CODE> (--execute) Specify code to rename file (\$_), e.g.
-     's/\.old\$/\.bak/'
  -D <NAME> (--delete) Delete scriptlet
  -d  (--dry-run) Dry-run (implies -v)
+ -e <CODE> (--execute) Specify Perl code to rename file (\$_). Can be specified
+     multiple times.
  -f  (--files) Only process files, do not process directories
  -h  (--help) Show this help
  -l  (--list) list all scriptlets
- -M <MODE> (--mode) Specify mode, default is 'rename' (or 'r'). Use
-     'copy' or 'c' to copy instead of rename, 'symlink' or 's' to create
-     a symbolic link, and 'link' or 'l' to create a (hard) link.
- -o  (--overwrite) Overwrite (by default, ".1", ".2", and so on will be
-     appended to avoid overwriting existing files)
+ -M <MODE> (--mode) Specify mode, default is 'rename' (or 'r'). Use 'copy' or 'c'
+     to copy instead of rename, 'symlink' or 's' to create a symbolic link, and
+     'link' or 'l' to create a (hard) link.
+ -o  (--overwrite) Overwrite (by default, ".1", ".2", and so on will be appended
+     to avoid overwriting existing files)
  -p  (--parents) Create intermediate directories
  -R  (--recursive) Recursive
- -r  (--reverse) reverse order of processing (by default order is
-     asciibetically)
+ -r  (--reverse) reverse order of processing (by default order is asciibetically)
  -S  (--no-symlinks) Do not process symlinks
  -s <NAME> (--show) Show source code for scriptlet
  -V  (--version) Print version and exit
  -v  (--verbose) Verbose
  -w <NAME> (--write) Write code specified in -e as scriptlet
+ -x <NAME> Execute a scriptlet. Can be specified multiple times. -x is optional
+     if there is only one scriptlet to execute, and scriptlet name is specified
+     as the first argument, and there is no -e specified.
 
 USAGE
 
@@ -203,7 +227,7 @@ sub load_scriptlet {
 
 sub load_scriptlets {
     my ($self) = @_;
-    $self->{'scriptlets'} ||= $self->find_scriptlets();
+    $self->{'scriptlets'} //= $self->find_scriptlets();
 }
 
 sub find_scriptlets {
@@ -213,14 +237,14 @@ sub find_scriptlets {
     eval { require App::perlmv::scriptlets::std };
     if (%App::perlmv::scriptlets::std::scriptlets) {
         $res->{$_} = { code => $App::perlmv::scriptlets::std::scriptlets{$_},
-                       from => "App::perlmv::scriptlets::std.pm" }
+                       from => "App::perlmv::scriptlets::std" }
             for keys %App::perlmv::scriptlets::std::scriptlets;
     }
 
     eval { require App::perlmv::scriptlets };
     if (%App::perlmv::scriptlets::scriptlets) {
         $res->{$_} = { code => $App::perlmv::scriptlets::scriptlets{$_},
-                       from => "App::perlmv::scriptlets.pm" }
+                       from => "App::perlmv::scriptlets" }
             for keys %App::perlmv::scriptlets::scriptlets;
     }
 
@@ -284,8 +308,7 @@ sub delete_user_scriptlet {
 }
 
 sub compile_code {
-    my ($self) = @_;
-    my $code = $self->{'code'};
+    my ($self, $code) = @_;
     no strict;
     no warnings;
     local $_ = "-TEST";
@@ -296,8 +319,7 @@ sub compile_code {
 }
 
 sub run_code_for_cleaning {
-    my ($self) = @_;
-    my $code = $self->{'code'};
+    my ($self, $code) = @_;
     no strict;
     no warnings;
     local $_ = "-CLEAN";
@@ -307,70 +329,89 @@ sub run_code_for_cleaning {
 }
 
 sub run_code {
-    my ($self) = @_;
-    my $code = $self->{'code'};
+    my ($self, $code) = @_;
     no strict;
     no warnings;
+    my $orig_ = $_;
     local $App::perlmv::code::TESTING = 0;
     local $App::perlmv::code::COMPILING = 0;
-    my $orig_ = $_;
     # It does need a package declaration to run it in App::perlmv::code
     my $res = eval "package App::perlmv::code; $code";
     die "FATAL: Code doesn't compile: code=$code, errmsg=$@\n" if $@;
     if (defined($res) && length($res) && $_ eq $orig_) { $_ = $res }
 }
 
+sub _sort {
+    my $self = shift;
+    $self->{reverse_order} ? (reverse sort @_) : (sort @_);
+}
+
 sub process_items {
-    my ($self, @items) = @_;
-    @items = $self->{'reverse_order'} ? (reverse sort @items) : (sort @items);
-    for my $item (@items) {
-        next if !$self->{'process_symlink'} && (-l $item);
-        if (-d _) {
+    my ($self, $code, $code_is_final, $items) = @_;
+    my $i = 0;
+    while ($i < @$items) {
+        my $item = $items->[$i];
+        $i++;
+        if ($item->{cwd}) {
+            chdir $item->{cwd} or die "Can't chdir to `$item->{cwd}`: $!";
+        }
+        next if !$self->{'process_symlink'} && (-l $item->{real_name});
+        if (-d $item->{real_name}) {
             next unless $self->{'process_dir'};
             if ($self->{'recursive'}) {
                 my $cwd = getcwd();
-                if (chdir $item) {
-                    print "INFO: chdir `$cwd/$item` ...\n" if $self->{'verbose'};
+                if (chdir $item->{real_name}) {
+                    print "INFO: chdir `$cwd/$item->{real_name}` ...\n"
+                        if $self->{'verbose'};
                     local *D;
                     opendir D, ".";
-                    my @d = grep { $_ ne '.' && $_ ne '..' } readdir D;
+                    my @subitems =
+                        $self->_sort(
+                            map { {name_for_script => $_, real_name => $_} }
+                            grep { $_ ne '.' && $_ ne '..' }
+                                readdir D
+                            );
                     closedir D;
-                    $self->process_items(@d);
+                    $self->process_items($code, $code_is_final, \@subitems);
+                    splice @$items, $i-1, 0, @subitems;
+                    $i += scalar(@subitems);
+                    $subitems[0]{cwd} = "$cwd/$item->{real_name}";
                     chdir $cwd or die "FATAL: Can't go back to `$cwd`: $!\n";
                 } else {
-                    warn "WARN: Can't chdir to `$cwd/$item`, skipped\n";
+                    warn "WARN: Can't chdir to `$cwd/$item->{real_name}`, ".
+                        "skipped\n";
                 }
             }
         }
-        $self->process_item($item, \@items);
+        $self->process_item($code, $code_is_final, $item, $items);
     }
 }
 
 sub process_item {
-    my ($self, $filename, $items) = @_;
+    my ($self, $code, $code_is_final, $item, $items) = @_;
 
-    local $_ = $filename;
-    $App::perlmv::code::FILES = $items;
-    my $old = $filename;
-    $self->run_code();
+    $App::perlmv::code::FILES =
+        [map {ref($_) ? $_->{name_for_script} : $_} @$items];
+    local $_ = $item->{name_for_script};
+    my $old = $item->{real_name};
+    $self->run_code($code);
     my $new = $_;
 
-    return if abs_path($old) eq abs_path($new);
+    my $aold = abs_path($old);
+    #die "Invalid path $old" unless defined($aold);
+    $aold = "" if !defined($aold);
+    my $anew = abs_path($new);
+    #die "Invalid path $new" unless defined($anew);
+    $anew = "" if !defined($anew);
+    $self->{_exists}{$aold}++ if (-e $aold);
+    return if $aold eq $anew;
 
-    my $cwd = getcwd();
-    my $orig_new = $new;
-    unless ($self->{'overwrite'}) {
-        my $i = 1;
-        while (1) {
-            if ((-e $new) || exists $self->{_exists}{"$cwd/$new"}) {
-                $new = "$orig_new.$i";
-                $i++;
-            } else {
-                last;
-            }
-        }
-        $self->{_exists}{"$cwd/$new"}++;
+    $item->{name_for_script} = $new;
+    unless ($code_is_final) {
+        push @{ $item->{intermediates} }, $new;
+        return;
     }
+
     my $action;
     if (!defined($self->{mode}) || $self->{mode} =~ /^(rename|r)$/) {
         $action = "rename";
@@ -384,8 +425,26 @@ sub process_item {
         die "Unknown mode $self->{mode}, please use one of: ".
             "rename (r), copy (c), symlink (s), or link (l).";
     }
+
+    my $orig_new = $new;
+    unless ($self->{'overwrite'}) {
+        my $i = 1;
+        while (1) {
+            if ((-e $new) || exists $self->{_exists}{$anew}) {
+                $new = "$orig_new.$i";
+                $anew = abs_path($new);
+                $i++;
+            } else {
+                last;
+            }
+        }
+    }
+    $self->{_exists}{$anew}++;
+    delete $self->{_exists}{$aold} if $action eq 'rename';
     print "DRYRUN: " if $self->{dry_run};
-    print "$action `$old` -> `$new`\n" if $self->{verbose};
+    print "$action " . join(" -> ",
+        map {"`$_`"} $old, @{ $item->{intermediates} // []}, $new)."\n"
+            if $self->{verbose};
     unless ($self->{dry_run}) {
         my $res;
 
@@ -440,17 +499,28 @@ sub rename {
         @items  = @{ $self->{'items'} // [] };
     }
 
-    # for cleaning between run
+    @items = map { {real_name=>$_, name_for_script=>$_} } $self->_sort(@items);
+
     if ($self->{_compiled}) {
-        $self->run_code_for_cleaning();
-    } else {
-        $self->compile_code();
-        $self->{_compiled} = 1;
+        # another run, clean first
+        $self->run_code_for_cleaning($_) for @{ $self->{'codes'} };
     }
     $self->{_exists} = {};
-    $self->process_items(@items);
+    local $self->{'recursive'} = $self->{'recursive'};
+    for (my $i=0; $i < @{ $self->{'codes'} }; $i++) {
+        my $code = $self->{'codes'}[$i];
+        $self->compile_code($code) unless $self->{'compiled'};
+        next if $self->{'check'};
+        my $code_is_final = ($i == @{ $self->{'codes'} }-1);
+        $self->{'recursive'} = 0 if $i;
+        #print STDERR "DEBUG: items (before): ".dump(@items)."\n";
+        $self->process_items($code, $code_is_final, \@items);
+        #print STDERR "DEBUG: items (after): ".dump(@items)."\n";
+    }
+    $self->{'compiled'}++;
 }
 
+
 1;
 
 __END__
@@ -462,13 +532,22 @@ App::perlmv - Rename files using Perl code.
 
 =head1 VERSION
 
-version 0.37
+version 0.38
+
+=head1 DESCRIPTION
+
+This application is used via the B<perlmv> command. See L<perlmv> for more
+documentation.
 
 =for Pod::Coverage .+
 
+=head1 SEE ALSO
+
+L<perlmv>
+
 =head1 AUTHOR
 
-  Steven Haryanto <stevenharyanto@gmail.com>
+Steven Haryanto <stevenharyanto@gmail.com>
 
 =head1 COPYRIGHT AND LICENSE
 
